package uk.co.hcsoftware.commander.bookmarks;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;
import uk.co.hcsoftware.commander.Main;
import uk.co.hcsoftware.commander.Url;

import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeModel;
import javax.xml.parsers.*;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.beans.XMLDecoder;
import java.beans.XMLEncoder;
import java.io.*;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.prefs.Preferences;


public class BookmarkReader {
    public static final Logger log = Logger.getLogger(BookmarkReader.class.getName());

    //private static final Preferences p = Preferences.userNodeForPackage(BookmarksBar.class);

    private static BookmarkReader instance = new BookmarkReader();

    public static BookmarkReader getInstance(){
        return instance;
    }

    private BookmarkReader() {
    }
    private File bookmarksFile(){
        //by default bookmarks go in ${HOME}.commanderbond/bookmarks.xml
        File homeDir = new File(System.getProperty("user.home"));
        File commanderDir = new File(homeDir,".commanderbond");
        return new File(commanderDir,"bookmarks.xml");
    }
    public void saveBookmarks(TreeModel bookmarks) {
        //  p.put("bookmarks",write(bookmarks));
        try {
            writeBookmarks(bookmarks,bookmarksFile());
        } catch (ParserConfigurationException e) {
            log.log(Level.WARNING, "couldn't save bookmarks", e);
        } catch (TransformerException e) {
            log.log(Level.WARNING, "couldn't save bookmarks", e);
        } catch (IOException e) {
            log.log(Level.WARNING, "couldn't save bookmarks", e);
        }
    }
    private void addLocationElement(Document d, Element parent, Url url, String target){
        Element e = d.createElement("location");
        e.setAttribute("url",url.stringify());
        e.setAttribute("target",target);
        parent.appendChild(e);
    }
    /**
     * prior to 20051024 bookmarks were stored in the user's preference tree under the rootpackagename
     * This was changed to store them in a file under the user's home dir for the following reasons:
     * <ol><li>make it easy for the user to copy bookmarks from one machine to another, (e.g. share them with other users) and
     * edit them through means other than the GUI tools provided</li>
     * <li>preparation for anticipated expansion of bookmarks system (they will be more complicated, allow saving of sort,filter prefs etc.)
     * by having a more elaborate storage format than default javabean/xml encoding</li>
     * </ol>
     * The old changeover should be as invisible as possible to the user. E.g. always write in the new format, try to read
     * in the new format - if fails try to read in the old<br/>
     *
     * @return
     * @param addSaveListener
     */
    public DefaultTreeModel readBookmarks(boolean addSaveListener) {
        DefaultTreeModel dtm;
        try {

            //first look for a file and try to read from that
            File f = bookmarksFile();
            if(f.exists()){//read from the boomarks file
                dtm= readBookmarks(f);
            }
            else{ //else try to read from prefs (legacy)
                Preferences p = Preferences.userNodeForPackage(Main.class);
                final String bookmarkString = p.get("bookmarks","");
                dtm=  convertList(read(bookmarkString));
            }
        } catch (Exception e) {
            log.log(Level.WARNING,"read bms",e);
            final List<Bookmark> l = Collections.emptyList();
            dtm= convertList(l);
        }
        if(addSaveListener){
            final DefaultTreeModel dtm1 = dtm;
            dtm.addTreeModelListener(new TreeModelListener() {
                private void save(){
                    log.fine("saving bookmarks in response to treechagne");
                    saveBookmarks(dtm1);
                }
                public void treeNodesChanged(TreeModelEvent e) { save(); }
                public void treeNodesInserted(TreeModelEvent e) { save(); }
                public void treeNodesRemoved(TreeModelEvent e) { save(); }
                public void treeStructureChanged(TreeModelEvent e) { save(); }
            });
        }
        return dtm;
    }
    private DefaultTreeModel convertList(List<Bookmark> l){
        DefaultMutableTreeNode root = new DefaultMutableTreeNode();
        for(Bookmark b:l){
            root.add(new DefaultMutableTreeNode(b,true));
        }
        return new DefaultTreeModel(root);
    }

    private DefaultTreeModel readBookmarks(File fromFile) throws IOException, ParserConfigurationException, SAXException {
        SAXParser sp = SAXParserFactory.newInstance().newSAXParser();
        InputStream is = new BufferedInputStream(new FileInputStream(fromFile));
        //final List<Bookmark> bookmarks = new ArrayList<Bookmark>();
        //map of path (e.g. path+sep+name)
        try {

            final DefaultMutableTreeNode root = new DefaultMutableTreeNode(new Bookmark(""));
            final DefaultTreeModel dtm = new DefaultTreeModel(root);

            sp.parse(is,new DefaultHandler(){
                Bookmark b;
                DefaultMutableTreeNode parent=root;

                public void endElement(String uri, String localName, String qName) {
                    if("directory".equals(qName)){
                        parent=(DefaultMutableTreeNode) parent.getParent();
                        if(parent==null) parent=root;
                    }
                }

                public void startElement(String uri, String localName, String qName, Attributes attributes) {
                    if("directory".equals(qName)){
                        b=new Bookmark();
                        final String name = attributes.getValue("name");
                        b.setName(name);
                        final DefaultMutableTreeNode node = new DefaultMutableTreeNode(b);
                        parent.add(node);
                        parent=node;
                    }
                    else if("bookmark".equals(qName)){
                        b=new Bookmark();
                        final String name = attributes.getValue("name");
                        b.setName(name);
                        parent.add(new DefaultMutableTreeNode(b));
                    }
                    else if("location".equals(qName)){
                        String target = attributes.getValue("target");
                        try {
                            if("left".equals(target)){
                                b.setF1(Url.getUrl(attributes.getValue("url")));
                            }
                            else if("right".equals(target)){
                                b.setF2(Url.getUrl(attributes.getValue("url")));
                            }
                        } /*catch (MalformedURLException e) {
                            log.log(Level.WARNING, "couldn't parse urlval="+attributes.getValue("url"), e);
                        }*/ catch (NullPointerException e){
                            log.warning("encountered location el but bookmark was null!");
                        }
                    }
                }
            });
            return dtm;
        } finally {
            is.close();
        }
    }
    private void writeBookmarks(final TreeModel bookmarks, File toFile) throws ParserConfigurationException, IOException, TransformerException {
        DocumentBuilder db = DocumentBuilderFactory.newInstance().newDocumentBuilder();
        final Document d = db.newDocument();
        final Element rootXmlElement = d.createElement("bookmarks");
        d.appendChild(rootXmlElement);
        //recurse down tree writing bookmarks
        Runnable treeWriter = new Runnable() {
            public void run() {
                 DefaultMutableTreeNode rootDmtn = (DefaultMutableTreeNode) bookmarks.getRoot();
                 if(rootDmtn!=null){
                     for (int i=0;i<rootDmtn.getChildCount();i++){
                         DefaultMutableTreeNode child = (DefaultMutableTreeNode) rootDmtn.getChildAt(i);
                         writeNode(rootXmlElement, child);
                     }
                 }
            }

            private void writeNode(Element parentXmlElement, DefaultMutableTreeNode dmtn){
                //store name, url1, url2
                //url1 and url2 are optional
                //something like <bookmarks><bookmark name="" path=""/><location url="" target="left"/><location url="" target="right"/>
                // <directory name="hello"><bookmark/> <bookmark/></directory>
                // </bookmarks>
                //example <bookmark name="mybookmark" parentpath="mybookmarks"/> <bookmark name="mybookmarks" parentpath=""/>
                //in future maybe: <bookmarks><bookmark name="" path=""/></bookmarks>
                Bookmark b = (Bookmark) dmtn.getUserObject();
                if(!dmtn.isLeaf()){ //recurse
                    Element bkEl=d.createElement("directory");
                    bkEl.setAttribute("name",b.getName());
                    parentXmlElement.appendChild(bkEl);
                    for (int i=0;i<dmtn.getChildCount();i++){
                        DefaultMutableTreeNode child = (DefaultMutableTreeNode) dmtn.getChildAt(i);
                        writeNode(bkEl, child);
                    }
                }
                else{  //must be a bookmark node
                    Element bkEl=d.createElement("bookmark");
                    parentXmlElement.appendChild(bkEl);
                    bkEl.setAttribute("name",b.getName());
                    final Url url1val = b.getF1();
                    if (url1val!=null) {
                        addLocationElement(d,bkEl,url1val,"left");
                    }
                    final Url url2val = b.getF2();
                    if (url2val!=null) {
                        addLocationElement(d,bkEl,url2val,"right");
                    }
                }
            }

        };
        treeWriter.run();

//        for(Bookmark b:bookmarks){
//            //store name, url1, url2
//            //url1 and url2 are optional
//            //something like <bookmarks><bookmark name="" path="/"/><location url="" target="left"/><location url="" target="right"/></bookmarks>
//            //path is the path to the parent e.g. /mybookmarks so the node's path is /mybookmarks/mynode
//            //directories are stored like <bookmark path="/hello/"/> where the path is the path to parent
//            //in future maybe: <bookmarks><bookmark name="" path=""/></bookmarks>
//            Element bkEl=d.createElement("bookmark");
//            rootXmlElement.appendChild(bkEl);
//            Attr a = d.createAttribute("name");
//            a.setValue(b.getName());
//            bkEl.setAttribute("name",b.getName());
//            bkEl.setAttribute("path","/");
//            final URL url1val = b.getF1();
//            if (url1val!=null) {
//                addLocationElement(d,bkEl,url1val,"left");
//            }
//            final URL url2val = b.getF2();
//            if (url2val!=null) {
//                addLocationElement(d,bkEl,url2val,"right");
//            }
//        }
        //now write to file

        Transformer t = TransformerFactory.newInstance().newTransformer();

        //create the file if necessary
        if(!toFile.exists()){
            final File parent = toFile.getParentFile();
            log.fine("creating parent="+parent);
            log.fine("result="+parent.mkdirs());
        }
        else log.finer("parent exists");
        final BufferedOutputStream os = new BufferedOutputStream(new FileOutputStream(toFile));
        final DOMSource source = new DOMSource(d.getDocumentElement());
        t.transform(source,new StreamResult(os));
        os.close();
    }

    private List<Bookmark> read(String s){
        if (s!=null && s.trim().length()>0) {
            try {
                log.finer("decode s="+s);
                XMLDecoder d = new XMLDecoder(new ByteArrayInputStream(s.getBytes("UTF8")));
                final List<Bookmark> bookmarks = new ArrayList<Bookmark>();
                try {
                    while(true){
                        final uk.co.hcsoftware.commander.Bookmark.BookmarkBean b = (uk.co.hcsoftware.commander.Bookmark.BookmarkBean) d.readObject();
                        if(b!=null){
                            bookmarks.add(b.createBookmark());
                        }
                    }
                } catch (ArrayIndexOutOfBoundsException e) {
                    ;  //not sure how else to determine when there are no more objects
                }
                d.close();
                log.fine("read bookmarks="+bookmarks);
                return bookmarks;
            } catch (UnsupportedEncodingException e) {
                log.log(Level.WARNING,"unsupported encoding utf8",e);
            } catch (ClassCastException e) {
                log.log(Level.WARNING,"ClassCastException",e);
            } catch(Exception e){
                log.log(Level.WARNING,"Exception",e);
            }
        }
        return new ArrayList<Bookmark>();
    }

    //this is the old write method
    private String write(List<Bookmark> bookmarks){
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        XMLEncoder e = new XMLEncoder(baos);
        for (Bookmark b:bookmarks) {
            e.writeObject(b.getBean());
        }
        e.close();
        try {
            final String s = new String(baos.toByteArray(),"UTF8");
            log.finer("wrote bookmarks="+s);
            return s;
        } catch (UnsupportedEncodingException e1) {
            log.log(Level.WARNING,"unsupported encoding utf8",e1);
        }
        return null;
    }
}
