package uk.co.hcsoftware.commander.gui;

import uk.co.hcsoftware.commander.*;
import uk.co.hcsoftware.commander.tree.CommanderTree;
import uk.co.hcsoftware.commander.treeplugin.TreePlugin;

import javax.swing.*;
import javax.swing.border.Border;
import java.awt.*;
import java.awt.event.*;
import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.prefs.Preferences;

/**
 * todo: treeplugin: findresults bookmarks? clearcase
 * DirecotryView manages a url field and a directory-view based on the url field
 */

public class DirectoryView extends JPanel implements Commandable, HostFrame,Themeable{
    private static final Logger log = Logger.getLogger(DirectoryView.class.getName());
    private static final Preferences prefs = Preferences.userNodeForPackage(Main.class);

    private JTextField locationField=new JTextField();

    private final String paneName;
    private final CommanderFrame commanderFrame;
    private final CommanderFocus cf;
    private JLabel statusLabel = new JLabel();


    private static final int GAP = 6;
    private final int borderWidth = GAP/2;
    private TreePlugin tp;
    private Border emptyBorder;
    // private java.util.List<CtreeModel.ChangeListener<ZipEntry>> changeListeners=new ArrayList<CtreeModel.ChangeListener<ZipEntry>>();


    public void updateStatus(final String status) {
        if(SwingUtilities.isEventDispatchThread()){
            setStatusText(status);
        }
        else{
            SwingUtilities.invokeLater(new Runnable(){
                public void run() {
                    setStatusText(status);
                }
            });
        }
    }

    private CommanderTree ct;

    public DirectoryView(String paneName, CommanderFrame mf, final CommanderFocus cf) {
        super(new BorderLayout(GAP,GAP));
        this.paneName = paneName;
        this.commanderFrame = mf;
        this.cf = cf;

        statusLabel.setPreferredSize(new Dimension(100,18));
        statusLabel.setBorder(BorderFactory.createEmptyBorder(2,2,2,2));

        final Border emptyBorder = BorderFactory.createEmptyBorder(GAP,GAP,GAP,GAP);
        setBorder(emptyBorder);

        final JScrollPane locSroll = new JScrollPane(locationField);
        locSroll.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_ALWAYS);

        add(locSroll,BorderLayout.NORTH);
        locationField.setAutoscrolls(true);

        add(statusLabel,BorderLayout.SOUTH);

        this.emptyBorder = emptyBorder;
        setBorder(this.emptyBorder);
        this.cf.addListener(new CommanderFocus.CommanderFocusListener() {
            public void focusChanged(Component lostFocus, Component gainedFocus) {
                setFocusColors();
            }
        });
        new ThemeListener(this);
        setupActions();
        locationField.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                try {
                    openUrl(locationField.getText());
                } catch (CannotOpenUrlException e1) {
                    log.log(Level.WARNING,"couldn't open url",e1);
                    JOptionPane.showMessageDialog(DirectoryView.this,"Cannot open that location:\n "+e1.getMessage(),"Open",JOptionPane.ERROR_MESSAGE);
                }
            }
        });
    }

    private void setFocusColors() {
        boolean hasFocus = cf.hasFocus(this);
        if(!hasFocus){
            log.finest("lost focus ");
            locationField.setBackground(backColor);
            locationField.setForeground(foreColor);
            setBorder(emptyBorder);
        }
        else{log.finest("ct gained focus");
            setBorder(focusBorder);
            locationField.setBackground(selColor);
            locationField.setForeground(backColor);
            if (ct!=null) {
                final boolean b = ct.requestFocusInWindow();
                log.finest("requestFocusInWindow="+b);
                if(!b){
                    log.finest("reqFoc=");
                    ct.requestFocus();
                }
            }
        }
    }


    public void setTreePlugin(final TreePlugin tp){
        //create a new tree
        tp.install(this);
        CommanderTree newTree = tp.createTree();
        if(ct!=null) remove(ct);
        if(this.tp!=null) tp.uninstall(this);
        add(newTree);
        this.ct=newTree;
        this.tp=tp;

        //manage focus
        ct.addFocusListener(new FocusAdapter() {
            public void focusGained(FocusEvent e) {
                cf.giveFocus(DirectoryView.this);
            }
        });

        //update the locationfield when the currentnode changes
        ct.addNodeListener(new CommanderTree.NodeListener() {
            public void parentNodeChanged(Object newParent, Object oldParent) {
                log.finest("node changed");
                final Url url = tp.getUrl(newParent);
                locationField.setText(url.getFriendlySpec());
                previousUrl=tp.getUrl(oldParent);
                prefs.put(paneName+".vfile",url.stringify());
            }
        });

        //forward up/down keystrokes to the commandertree
        ActionMap am = locationField.getActionMap();
        InputMap im = locationField.getInputMap(JComponent.WHEN_FOCUSED);
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_UP,0), "up");
        im.put(KeyStroke.getKeyStroke(KeyEvent.VK_DOWN,0), "down");
        am.put("up",ct.getForwardAction(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT,KeyStroke.getKeyStroke(KeyEvent.VK_UP,0)));
        am.put("down",ct.getForwardAction(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT,KeyStroke.getKeyStroke(KeyEvent.VK_DOWN,0)));

        repaint();
    }

    public boolean hasCommanderFocus(){
        return cf.hasFocus(this);
    }

    private Border focusBorder;
    private Border createFocusBorder(Color c) {
        return BorderFactory.createCompoundBorder(
                BorderFactory.createEmptyBorder(borderWidth,borderWidth,borderWidth,borderWidth),
                BorderFactory.createLineBorder(c,borderWidth)
        );
    }
    //Themeable interface
    private Color backColor,foreColor,selColor;

    public void setColors(Color backColor, Color foreColor, Color selColor) {
        focusBorder =createFocusBorder(selColor);
        this.backColor=backColor;
        this.foreColor=foreColor;
        this.selColor=selColor;
        setFocusColors();
        repaint();
    }

    public boolean getFocus(){
        if(inputBlocked){
            log.log(Level.FINEST,"ignoring reqeustfocus as input is blocked",new Throwable());
            return false;
        }
        if (ct!=null) {
            final boolean b = ct.requestFocusInWindow();
            log.finest("requestFocusInWindow="+b);
            if(!b) ct.requestFocus();
            return b;
        } else {
            return false;
        }
    }

    Action selectLocationBarAction=new AbstractAction() {
        public void actionPerformed(ActionEvent e) {
            locationField.selectAll();
            locationField.requestFocus();
        }
    };


    public String toString() {
        return "DirectoryView name="+paneName;
    }

    CancelledCallback cc = new CancelledCallback();

    public CancelledCallback getCancelledCallback() {
        return cc;
    }

    //Url getTargetUrl();
    public TreePlugin getTargetTreePlugin() {
        return commanderFrame.getTargetTreePlugin();
    }

    /**
     * block gui input. If ESC is pressed cc will be set to true (at which point you might want to abort what you're
     * doing and call unblockinput;
     * @param text  adds "(ESC to cancel)" to this
     * @param cc
     */
    public void blockInput(String text,final CancelledCallback cc) {
        if(cc!=null){
            cc.reset();
            blockednessEscapable=true;
            commanderFrame.blockInput(text+"\n\n"+"(ESC to cancel)",new KeyAdapter() {
                public void keyPressed(KeyEvent e) {
                    if(KeyEvent.VK_ESCAPE==e.getKeyCode()){
                        cc.cancel();
                    }
                }
            });
            inputBlocked=true;
        }
        else{
            blockednessEscapable=false;
            commanderFrame.blockInput(text);
            inputBlocked=true;
        }
    }
    private boolean blockednessEscapable=false;
    private boolean inputBlocked=false;
    /**
     * block GUI input (e.g. keys, mouse) by putting a glasspane over the whole frame
     * @param block if false remove the current glasspane
     * @param text to display on the glasspane
     */
    public void blockInput(boolean block, String text) {
        blockednessEscapable=false;
        log.fine("blockinput block="+block+" text="+text);
        if (block) {
            commanderFrame.blockInput(text);
            inputBlocked=true;
        }
        else{
            commanderFrame.unblockInput();
            inputBlocked=false;
        }
    }

    /**
     * call in EDT
     * @param text
     */
    protected void setStatusText(String text){
        commanderFrame.updateBlockedMessage(blockednessEscapable?text+"\n\n"+"(ESC to cancel)":text);
    }

    private void setupActions() {
        final InputMap inputMap = getInputMap(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
        final ActionMap actionMap = getActionMap();

        java.util.List<Action> actions=new ArrayList<Action>();

        actions.add(ActionHelper.setupAction(selectLocationBarAction,KeyStroke.getKeyStroke(KeyEvent.VK_L,Main.ctrl_mask),"lbar"));
        actions.add(ActionHelper.setupAction(new AbstractAction("delete") {
            public void actionPerformed(ActionEvent e) {
                delete();
            }
        }, KeyStroke.getKeyStroke(KeyEvent.VK_DELETE,0),"delete"));
        inputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_BACK_SPACE,0),"delete");

        for(Action a:actions){
            inputMap.put((KeyStroke) a.getValue(Action.ACCELERATOR_KEY),a.getValue(Action.ACTION_COMMAND_KEY));
            actionMap.put(a.getValue(Action.ACTION_COMMAND_KEY),a);
        }
    }

    public void delete() {
        if(tp!=null) tp.delete();
    }

    public void launch() {
        if(tp!=null) tp.launch();
    }

    public void mkdir() {
        if(tp!=null) tp.mkdir();
    }

    public void move() {
        if(tp!=null) tp.move();
    }

    public void rename() {
        if(tp!=null) tp.rename();
    }

    public void copy() {
        if(tp!=null) tp.copy();
    }

    public void duplicate() {
        if(tp!=null) tp.duplicate();
    }

    public void openUrl(Url url) throws CannotOpenUrlException{
        if(url==null){
            log.warning("ignoring null url");
            return;
        }
        if (tp!=null && Url.isRecommendedPlugin(tp,url)) {
            log.fine("opening url in current plugin");
            openUrl(tp, url);
        } else {
            //todo store current state so we can restore it if we exit this plugin?
            try {
                log.fine("url="+url+" not understood by this plugin");
                TreePlugin tp2=Url.getPlugin(url);
                setTreePlugin(tp2);
                log.fine("loaded new plugin="+tp2);
                locationField.requestFocus();
                openUrl(tp2,url);
            } catch (NoPluginAvailableException e) {
                throw new CannotOpenUrlException(url, "no plugin understands this URL");
            }
        }
    }


    private void openUrl(TreePlugin tp, Url pane1Url) throws CannotOpenUrlException {
        log.fine("open url="+pane1Url);
        tp.openUrl(pane1Url);
    }

    private Url previousUrl=null;
    private void openUrl(String text) throws CannotOpenUrlException {
        openUrl(Url.getUrl(text));

    }

    public Url getURL() {
        if (ct!=null) {
            final Object node = ct.getNode();
            if(node==null) return null;
            return tp.getUrl(node);
        } else {
            return null;
        }
    }

    public TreePlugin getTreePlugin() {
        return tp;
    }

    public void showPreviousUrl() throws CannotOpenUrlException {
        log.fine("showprevurl="+previousUrl);
        if(previousUrl!=null) openUrl(previousUrl);
    }


    public void refresh() {
        if(ct!=null) ct.setNodeAsynchronous(ct.getNode(), null, -1, false, null);
    }

    public Url getTargetUrl() {
        return commanderFrame.getTargetUrl();
    }


    public void setStatus(String status) {
        log.finest("ssetstaus text="+status);
        statusLabel.setText(status);
    }
    public String getStatus(){
        return statusLabel.getText();
    }
}
