package uk.co.hcsoftware.commander.treeplugin.zip;

import uk.co.hcsoftware.commander.Url;

import java.util.*;
import java.util.logging.Logger;
import java.util.zip.ZipEntry;


/**
 * Title: ZipNode
 * Copyright:    Copyright (c) 29-Nov-2005
 *
 * @author spurlinn
 */
public class ZipNode{
    private static final Logger log = Logger.getLogger(ZipNode.class.getName());
    private ZipEntry ze;
    private boolean isDirectory;
    private final ZipNode parent;

    public ZipNode(ZipNode parent, String[] path) {
        this.parent=parent;
        this.path = path;
        isDirectory=true;
        if(parent!=null) parent.addChild(this);
    }

    private String[] getPath(String path) {
        final String[] strings = path.split("/");
        log.fine("path="+path+" yields="+Arrays.asList(strings));
        return strings;
    }

    public void setZipEntry(ZipEntry ze) {
        this.ze = ze;
        isDirectory=ze.isDirectory();
    }

    private final String[] path; //full path to node, including name
    public String getName(){
        if(path.length==0) return "";
        return path[path.length-1];
    }
    public ZipNode getChild(String name){
        if(children==null) return null;
        for(ZipNode zn:children){
            if(name.equals(zn.getName())) return zn;
        }
        return null;
    }
    public ZipNode getOrCreateChild(String name){
        ZipNode zn = getChild(name);
        if(zn==null){
            String[] newPath = new String[path.length+1];
            System.arraycopy(path,0,newPath,0,path.length);
            newPath[newPath.length-1]=name;
            zn = new ZipNode(this,newPath);
        }
        return zn;
    }
    /**
     *
     * @param zipPath e.g. plugins/net.sourceforge.eclipseccase.source_1.1.0/src/net.sourceforge.eclipseccase.ui_1.1.0/icons/full/hijacked_ovr.gif
     * @return
     */
    public ZipNode getOrCreateDescendant(String zipPath) {
        String[] path = getPath(zipPath);
        ZipNode curr=this;
        for(String childName:path){
            curr=curr.getOrCreateChild(childName);
        }
        return curr;
    }
    public ZipNode getDescendant(String zipPath) {
        String[] path = getPath(zipPath);
        ZipNode curr=this;
        for(String childName:path){
            curr=curr.getChild(childName);
            if(curr==null) return null;
        }
        return curr;
    }


    private List<ZipNode> children;
    public void addChild(ZipNode zn){
        if(children==null) children=new ArrayList<ZipNode>();
        children.add(zn);
    }
    public List<ZipNode> getChildren() {
        if(children==null) return Collections.emptyList();
        return children;
    }
    public boolean isDirectory() {
        return isDirectory;
    }
    public ZipNode getParent() {
        return parent;
    }

    public long getSize() {
        return ze==null?-1:ze.getSize();
    }

    public long getTime() {
        return ze==null?-1:ze.getTime();
    }
    public int getCompressedRatio() {
        if(ze==null) return -1;
        final long compressed_size = (ze.getCompressedSize()*100);
        final long original_size = (ze.getSize());
        //log.fine("comp="+compressed_size+" orig="+original_size);
        if(original_size==0 || original_size==-1) return -1;
        return 100-(int) (compressed_size/original_size);
    }

    public boolean hasEntry() {
        return ze!=null;
    }

    public ZipEntry getZipEntry() {
        return ze;
    }

    /**
     *
     * @return flat list of this node's zipentry (if it has one) followed by all its descendants' zipentries (if they have them)
     */
    public List<ZipEntry> getZipEntries() {
        LinkedList<ZipEntry> ll = new LinkedList<ZipEntry>();
        if(hasEntry()) ll.add(getZipEntry());
        for(ZipNode child:children){
            ll.addAll(child.getZipEntries());
        }
        return ll;
    }


    public Url getUrl(Url url) {
        //take everything up to the ! from this url, and add our path
        log.fine("getUrl url="+url);
        String base = url.stringify();
        int i=base.indexOf('!');
        if(i>-1){
            base=base.substring(0,i);
        }
        base+='!';
        log.fine("baseurl="+base);
        //now build the path to this node
        String relativePath = getPathFromRoot();
        log.fine("relativePath="+relativePath+" node="+toString());
        String fullUrl = base+relativePath;
        log.fine("fullURl="+fullUrl);
        return Url.getUrl(fullUrl);
    }

    /**
     * includes leading /
     * @return
     */
    private String getPathFromRoot() {
        StringBuffer sb = new StringBuffer();
        if(path.length==0) sb.append('/');
        for(String element:path){
            sb.append('/');
            sb.append(element);
        }
        return sb.toString();
    }

    public String toString() {
        return "name="+getName()+" path="+Arrays.asList(path);
    }
}
