package uk.co.hcsoftware.commander.treeplugin.zip;

import uk.co.hcsoftware.commander.*;
import uk.co.hcsoftware.commander.gui.ActionHelper;
import uk.co.hcsoftware.commander.gui.HostFrame;
import uk.co.hcsoftware.commander.tree.*;
import uk.co.hcsoftware.commander.treeplugin.*;
import uk.co.hcsoftware.commander.treeplugin.file.FileEventManager;

import javax.swing.*;
import java.awt.*;
import java.awt.datatransfer.*;
import java.awt.dnd.*;
import java.awt.event.*;
import java.io.*;
import java.net.MalformedURLException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipOutputStream;


/**
 *
 * todo: support for zipedit actions (mkdir etc) should overwrite zipfile with a new zipfile call writeZipFile and update model
 * todo: add 'create zip' action
 * todo: support bookmarking zips (and directories within zips) e.g. add url transferflavor
 * todo: detect changes to the underlying zipfile and update the view
 *
 * todo: think about having 'parent dir' for root of zip and disabling leftarrow (easy to click it by accident and exit the zip)
 *
 *
 * Title: ZipTreePlugin
 * Copyright:    Copyright (c) 28-Nov-2005
 *
 * @author spurlinn
 */
public class ZipTreePlugin implements TreePlugin{
    static final Logger log = Logger.getLogger(ZipTreePlugin.class.getName());
    private static final int MARK=0,NAME=1,SIZE=2,DATE=4,COMPRESSED=3; //attribute (column) ids

    private CommanderTree ct;
    private HostFrame hf;
    private ZipModel model = new ZipModel();
    private final SelectionMarkHandler<ZipNode> sml;// = new SelectionMarkHandler();
    private final ColumnSizeSaver css;
    private final ColorHandler ch;
    private Url url;
    private List<CtreeModel.ChangeListener<ZipNode>> changeListeners=new ArrayList<CtreeModel.ChangeListener<ZipNode>>();
    private File file;
    private ZipNode root;

    public ZipTreePlugin() {
        ct = new CommanderTree(model);
      // ct.setHideCurrentIfRoot(true);
        sml = new SelectionMarkHandler<ZipNode>(ct);
        css = new ColumnSizeSaver(ct,"mark_column_w","name_column_w","size_column_w","compress_column_w","date_column_w") ;
        ch=new ColorHandler(ct);
        ct.setColumnSpec(createColumnSpec());
        ct.setDefaultComparison(1,SortDirection.ascending);
        ct.setNameColumnIndex(1);
        ct.setDirColumnIndex(2);
        setupActions();
        setupDnD();
        // new StatusUpdater(ct,hf,sml);
        //ct.set
        sml.addSelectionMarkListener(new SelectionMarkHandler.SelectionMarkListener() {
            public void selectionMarkChanged(int[] selectedFiles) {
                List<ZipNode> nodes = new ArrayList<ZipNode>(selectedFiles.length);
                for(int i:selectedFiles){
                    nodes.add((ZipNode) ct.getNode(i));
                }
                fireNodesChanged(CtreeModel.ChangeListener.ChangeType.changed, nodes);
            }
        });
    }
    private void fireNodesChanged(CtreeModel.ChangeListener.ChangeType ct, java.util.List<ZipNode> nodes){
        for(CtreeModel.ChangeListener<ZipNode> cl: changeListeners){
            cl.nodesChanged(ct,nodes);
        }
    }


    final static Font plainFont = new Font("lucida sans",Font.PLAIN,12);
    final static Font italicPlainFont = new Font("lucida sans",Font.ITALIC /*| Font.BOLD*/,12);
    final static Font boldFont = new Font("lucida sans",Font.BOLD,12);
    private DefaultColumnSpec createColumnSpec() {
        DefaultColumnSpec cs = new DefaultColumnSpec(model);
        //marked,name,modified,size, ratio? packed?

        cs.setWidth(MARK,css.getColWidth(MARK,5));
        cs.setWidth(NAME,css.getColWidth(NAME,200));
        cs.setWidth(SIZE,css.getColWidth(SIZE,50));
        cs.setWidth(DATE,css.getColWidth(DATE,120));
        cs.setWidth(COMPRESSED,css.getColWidth(COMPRESSED,50));

        cs.setRenderer(MARK,new DefaultCommanderCellRenderer() { //mark column
            public Component getRendererComponent(CommanderTree ct, Object attributeValue, Object node, boolean isSelected, boolean hasFocus) {
                JLabel jl =  (JLabel)  super.getRendererComponent(ct
                        , attributeValue==null?"":((Boolean) attributeValue).booleanValue()?"*":""
                        , node, isSelected, hasFocus);
                setupLabel(node, jl, isSelected);
                return jl;
            }
        });
        cs.setRenderer(NAME,new DefaultCommanderCellRenderer() {  //name column
            public Component getSpecialRendererComponent(CommanderTree ct, SpecialValues value, Object node, boolean isSelected, boolean hasFocus) {
                JLabel jl =  (JLabel) super.getSpecialRendererComponent(ct, value, node, isSelected, hasFocus);
                jl.setFont(boldFont);
                ch.setColors(jl,isSelected, sml.isMarked((ZipNode) node));
                return jl;
            }

            public Component getRendererComponent(CommanderTree ct, Object attributeValue, Object node, boolean isSelected, boolean hasFocus) {
                JLabel jl =  (JLabel)  super.getRendererComponent(ct, attributeValue, node, isSelected, hasFocus);
                setupLabel(node, jl, isSelected);
                return jl;
            }
        })  ;
        cs.setRenderer(SIZE,new DefaultCommanderCellRenderer() { //size column
            NumberFormatter nf = new NumberFormatter();
            public Component getSpecialRendererComponent(CommanderTree ct, SpecialValues value, Object node, boolean isSelected, boolean hasFocus) {
                JLabel jl =  (JLabel) super.getSpecialRendererComponent(ct, value, node, isSelected, hasFocus);
                jl.setFont(boldFont);
                ch.setColors(jl,isSelected, sml.isMarked((ZipNode) node));
                return jl;
            }
            public Component getRendererComponent(CommanderTree ct, Object attributeValue, Object node, boolean isSelected, boolean hasFocus) {
                JLabel jl =  (JLabel)  super.getRendererComponent(ct
                        ,attributeValue==null?"": nf.formatBytes(((Long) attributeValue).longValue())
                        , node, isSelected, hasFocus);
                setupLabel(node, jl, isSelected);
                return jl;
            }
        });
        cs.setRenderer(DATE,new DefaultCommanderCellRenderer() { //date column
            private final DateFormat df = new SimpleDateFormat("dd-MMM-yy HH:mm");
            public Component getRendererComponent(CommanderTree ct, Object attributeValue, Object node, boolean isSelected, boolean hasFocus) {
                JLabel jl =  (JLabel)  super.getRendererComponent(ct
                        ,attributeValue==null?"": df.format(new Date((Long)attributeValue))
                        , node, isSelected, hasFocus);
                setupLabel(node, jl, isSelected);
                return jl;
            }
        });
        cs.setRenderer(COMPRESSED,new DefaultCommanderCellRenderer() { //size column
            public Component getRendererComponent(CommanderTree ct, Object attributeValue, Object node, boolean isSelected, boolean hasFocus) {
                final String av ;
                if(attributeValue==null){
                    av="";//?"": attributeValue+"%";
                }
                else{
                    Integer l = (Integer) attributeValue;
                    if(l.longValue()<0) av="";
                    else av=l.toString()+"%";
                }
                JLabel jl =  (JLabel)  super.getRendererComponent(ct
                        ,av
                        , node, isSelected, hasFocus);
                setupLabel(node, jl, isSelected);
                return jl;
            }
        });
        return cs;
    }
    private void setupLabel(Object node, JLabel jl, boolean isSelected) {
        ZipNode file = (ZipNode) node;
        boolean marked = false;
        if (file!=null) {
            jl.setFont(plainFont);
            if(file.isDirectory()){
                jl.setFont(boldFont);
            }
            else{
                jl.setFont(plainFont);
            }
            marked=sml.isMarked(file);
        }
        ch.setColors(jl,isSelected, marked);
    }





    public CommanderTree createTree() {
        return ct;
    }

    public void install(HostFrame o) {
        this.hf=o;
        new StatusUpdater(ct,hf,sml);
    }

    public void openUrl(Url url) throws CannotOpenUrlException {
        log.fine("open url="+url);
        this.url=url;
        String s = url.stringify();
      //  if(s.startsWith("zip:") || s.startsWith("jar:")){//assume it's a stringurl starting zip: or jar:
            String subUrl = s;//.substring(4);
            int i= subUrl.indexOf('!');
            String zipPath;
            if(i>-1 && i+1<subUrl.length()){
                zipPath=subUrl.substring(i+1);
                if(zipPath.startsWith("/")){
                    //since we are resolving relative to root, remove the leading /
                    zipPath=zipPath.substring(1);
                }
                subUrl=subUrl.substring(0,i);
            }
            else zipPath="";
            log.fine("subUrl="+subUrl);
            try {
                file = url.getFile();// new File(new URI(subUrl));
                ZipFile zf = new ZipFile(file);
                log.fine("zf entrycount="+zf.size());
                final ZipNode root = new ZipNode(null,new String[0]);
                //if it's a dir the name is the dirpath and isDirectory returns true
                //if it's a file the name is the filepath and isDirectory returns false
                for(Enumeration<? extends ZipEntry> en=zf.entries();en.hasMoreElements();){
                    final ZipEntry ze = en.nextElement();
                    if (!ze.isDirectory()) {
                        final String path = ze.getName();
                        ZipNode zn = root.getOrCreateDescendant(path);
                        zn.setZipEntry(ze);
                        if (log.isLoggable(Level.FINER)) {
                            log.finer("path="+path+" zn="+zn);
                        }
                    }
                }
                zf.close();
                this.root=root;
                ZipNode openNode = root.getDescendant(zipPath);
                if(openNode==null){
                    log.warning("couldn't open zipPath="+zipPath+" as it wasn't found!");
                    openNode=root;
                }
                ct.setNodeAsynchronous(openNode, null, -1, false, null);
            } catch (IOException e) {
                throw new CannotOpenUrlException(url, "error opening zip file",e);
            } catch (CannotConvertUrlToFileException e) {
                 throw new CannotOpenUrlException(url, "error opening zip file",e);
            }
     //   }
    }

    public Url getUrl(Object node) {
        ZipNode zn = (ZipNode) node;
        if(zn==null) return root.getUrl(url);
        return zn.getUrl(url);
    }

    public void save(Transferable t) throws NotSupportedException, CannotSaveException {
        throw new NotSupportedException();

    }

    public void uninstall(HostFrame hf) {

    }


    public void delete() {

    }
    public void launch() {
        int[] sel = sml.getSelectedOrMarked(true, true,true);
        //weed out directories and nodes without zipentries
        final ArrayList<ZipEntry> l = getZipEntries(sel, false, false);
        if(l.size()>0){
            launchFiles(l);
        }
        hf.getFocus();
    }
    FileLauncher fl = new FileLauncher();
    private void launchFiles(List<ZipEntry> l) {
        String s = l.size()==1?l.get(0).getName():l.size()+" files";
        String[] options = new String[]{"launch","cancel"};
        if ( 0== JOptionPane.showOptionDialog(ct
                ,"launch "+s+"?","Launch"
                ,JOptionPane.YES_NO_CANCEL_OPTION
                ,JOptionPane.QUESTION_MESSAGE,null
                ,options,options[0])) {
            try {
                List<File> files = writeTempFiles(l);
                fl.launchFiles(files);
            } catch (IOException e) {
                log.log(Level.WARNING,"ioe",e);
                JOptionPane.showMessageDialog(ct,"Error executing external process\n" +
                        "["+e.getMessage()+"]","Launch",JOptionPane.ERROR_MESSAGE);
            }
        }
    }

    public void mkdir() {
//        String dest=(String)JOptionPane.showInputDialog(ct,
//                "Create Directory:"
//                ,"Create Directory",JOptionPane.QUESTION_MESSAGE,null
//                ,null,"");
//        if(dest!=null && dest.trim().length()>0){
//            ZipNode zn = (ZipNode) ct.getNode();
//
//            Vfile d=new Vfile(getDirectory(),dest);
//            if(!d.exists() ){
//                if (d.mkdir()) {
//                    fem.fireEvent(new FilesChangedEvent(null,Collections.singletonList(d),null));
//                    //select the dir we just created
//                    selectFile(d);
//                } else {
//                    JOptionPane.showMessageDialog(ct
//                            ,"couldn't create directory"
//                            ,"Create Directory",JOptionPane.ERROR_MESSAGE);
//                }
//            } else{
//                JOptionPane.showMessageDialog(ct
//                        ,d+" already exists"
//                        ,"Create Directory"
//                        ,JOptionPane.WARNING_MESSAGE);
//                //select the existing dir
//                selectFile(d);
//            }
//            hf.getFocus();
//        }

    }

    public void move() {

    }

    public void rename() {

    }

    public void copy() {
        try {
            Transferable t = getTransferable();
            TreePlugin tp = hf.getTargetTreePlugin();
            if(t==null || tp==null){
                log.fine("ignoring copy as no transferable files or opposite was null");
                return;
            }
            else if(JOptionPane.YES_OPTION==JOptionPane.showConfirmDialog(ct,"Copy selected files to other pane?")){
                tp.save(t);
            }

        } catch (NotSupportedException e) {
            JOptionPane.showMessageDialog(ct,"The opposite pane doesn't understand my data!","Copy",JOptionPane.ERROR_MESSAGE);
        } catch (CannotSaveException e) {
            log.log(Level.WARNING, "exception", e);
            JOptionPane.showMessageDialog(ct,"Couldn't save those data\n"+e.getMessage(),"Copy",JOptionPane.ERROR_MESSAGE);
        }
    }

    public void duplicate() {

    }

    private class ZipModel implements CtreeModel<ZipNode>{
        public List<ZipNode> getChildren(ZipNode zipNode) {
            return zipNode.getChildren();
        }
        public boolean hasChildren(ZipNode ZipNode) {
            return ZipNode.isDirectory();
        }
        public ZipNode getParent(ZipNode ZipNode) {
              return ZipNode.getParent();
        }
        public int getAttributeCount() {
            return 5;
        }
        public String getAttributeName(int index) {
            switch(index){
                case MARK: return "";
                case NAME: return "Name";
                case SIZE: return "Size";
                case DATE: return "Date";
                default: return "Zipped to";
            }
        }
        public Object getAttributeValue(ZipNode ZipNode, int attribute) {
            switch(attribute){
                case MARK: return sml.isMarked(ZipNode);
                case NAME: return ZipNode.getName();
                case SIZE: return ZipNode.getSize();
                case DATE: return ZipNode.getTime();
                default: return ZipNode.getCompressedRatio();
            }
        }
        public boolean isComparable(int attribute) {
            return true;
        }

        private NameComparator nc=new NameComparator();
        private int compare(long l1, long l2){
            return l1<l2?-1:l1==l2?0:1;
        }
        private int compare(boolean b1, boolean b2){
            if(b1){
                if(b2) return 0;
                else return 1;
            }
            else if(b2) return -1;
            else return 0;
        }
        public int compare(ZipNode vfile, ZipNode vfile1, int attribute) {
            final int i;
            switch(attribute){
                case MARK:{
                    i=compare(sml.isMarked(vfile),sml.isMarked(vfile1));
                    break;
                }
                case NAME:
                    i= nc.compare(vfile.getName(),vfile1.getName());
                    break;
                case SIZE:
                    i=compare(vfile.getSize(),vfile1.getSize());
                    break;
                case DATE:
                    i=compare(vfile.getTime(),vfile1.getTime());
                    break;
                default:
                    i= compare(vfile.getCompressedRatio(),vfile1.getCompressedRatio());
                    break;
            }
            return i;
        }
        public SortDirection getDefaultSortDirection(int attribute) {
            switch(attribute){
                case MARK: return SortDirection.descending;
                case NAME: return SortDirection.ascending;
                case SIZE: return SortDirection.descending;
                case DATE: return SortDirection.descending;
                default: return SortDirection.ascending;
            }
        }

        public void addChangeListener(ChangeListener<ZipNode> changeListener) {
             changeListeners.add(changeListener);
        }

        public void removeChangeListener(ChangeListener<ZipNode> changeListener) {
             changeListeners.remove(changeListener);
        }
    }
    public List<ZipNode> getSelectedFiles(boolean includeSelected, boolean includeMarkedFiles, boolean includeParents){
        int[] rows = sml.getSelectedOrMarked(includeSelected, includeMarkedFiles,includeParents);
        List<ZipNode> l = new ArrayList<ZipNode>(rows.length);
        for(int ro:rows){
            l.add((ZipNode) ct.getNode(ro));
        }
        return l;
    }
    private void setupDnD() {
//        final DropTargetListener dtl = new DropTargetListener(){
//            public void dragEnter(DropTargetDragEvent evt) {
//                if(!evt.isDataFlavorSupported(DataFlavor.javaFileListFlavor)){
//                    log.finer("reject drag supported SEND_FLAVORS="+evt.getCurrentDataFlavorsAsList());
//                    evt.rejectDrag();
//                }
//                else evt.acceptDrag(DnDConstants.ACTION_COPY);
//            }
//            public void dragOver(DropTargetDragEvent evt) {; }
//            public void dragExit(DropTargetEvent evt) {;}
//            public void dropActionChanged(DropTargetDragEvent evt) {;}
//            public void drop(DropTargetDropEvent evt) {
//                onDrop(evt);
//            }
//        };
//        ct.setDropTarget(new DropTarget(ct,dtl));
        final DragSource ds = new DragSource();
        ct.createDefaultDragGestureRecognizer(DnDConstants.ACTION_COPY,new DragGestureListener() {
            public void dragGestureRecognized(DragGestureEvent dge) {
                onDrag(dge, ds);
            }
        },ds);
    }
    private static final DataFlavor[] SEND_FLAVORS={TreeStream.DATAFLAVOR, DataFlavor.javaFileListFlavor};
    /**
     *
     * @return a transferable object holding references (of various sorts) to all the selected files
     */
    private Transferable getTransferable(){
        //what we want here, is to recursively write all descendants of the selection, but only put the selected files/dirs in the transferable
        int[] sel = sml.getSelectedOrMarked(true, true,true);
        //weed out nodes without zipentries
        final ArrayList<ZipEntry> zes = getZipEntries(sel, true, true);
        final ArrayList<ZipEntry> roots = getZipEntries(sel, true, false);
        final List<ZipNode> znRoots = getSelectedFiles(true, true, true);
        log.fine(sel.length+" files selected");
        if(zes.isEmpty()){
            log.fine("ignore drag as sel.length==0");
            return null;
        }
        Transferable t = new Transferable() {
            public DataFlavor[] getTransferDataFlavors() {
                return SEND_FLAVORS;
            }
            public boolean isDataFlavorSupported(DataFlavor flavor) {
                for(DataFlavor f:SEND_FLAVORS) if(f.match(flavor)) return true;
                return false;
            }
            List<File> dragFilesList ;
            public Object getTransferData(DataFlavor flavor) throws UnsupportedFlavorException, IOException {
                final boolean dataFlavorSupported = isDataFlavorSupported(flavor);
                log.finer("getTransferData dataFlavorSupported="+dataFlavorSupported);
             //   if(!dataFlavorSupported) throw new UnsupportedFlavorException(flavor);
               // hf.blockInput(true,"extracting "+zes.size()+" files to temp location");
              //  ct.repaint();
                if(TreeStream.DATAFLAVOR.match(flavor)){
                    List<TreeStream.Node> nodeRoots= new ArrayList<TreeStream.Node>(roots.size());
                    final ZipFile zf = new ZipFile(file);
                    for(ZipNode ze:znRoots){
                        nodeRoots.add(new TreeStreamNode(zf,ze));
                    }
                    return new TreeStream(nodeRoots){
                        public void finishedReading() {
                            try {
                                zf.close();
                            } catch (IOException e) {
                                log.log(Level.WARNING, "exception", e);
                            }
                        }
                    };
                }
                else if (DataFlavor.javaFileListFlavor.match(flavor)) {
                    try {
                        if (dragFilesList==null) {
                            dragFilesList = new ArrayList<File>(zes.size());
                            writeTempFiles(zes);
                            dragFilesList.addAll(getTempFiles(roots));
                        }
                        return dragFilesList;
                    } catch (IOException e) {
                        log.log(Level.WARNING, "exception", e);
//showing an error message here seems to cause gui probs (message can't come to front but is blocking other input)
//                    JOptionPane.showMessageDialog(ct,"Couldn't extract file(s) from archive\n" +
//                            e.getMessage(),"Drag files",JOptionPane.ERROR_MESSAGE);
                        return Collections.emptyList();
                    }
                }
                else throw new UnsupportedFlavorException(flavor);
            }
        };
        return t;
    }

    private class TreeStreamNode implements TreeStream.Node{
        private final ZipFile zf;
        private ZipNode ze;
        public TreeStreamNode(ZipFile zf,ZipNode ze) {
            this.zf = zf;
            this.ze = ze;
        }

        public String getName() {
//            String name = ze.getName();
//            int i = name.lastIndexOf('/');
//            if(i>-1) name=name.substring(i+1);
//            log.fine("returning name="+name+" was="+ze.getName());
//            return name;
            return ze.getName();
        }

        public List<TreeStream.Node> getChildren() {
            if(ze.isDirectory()){
                List<TreeStream.Node> l = new ArrayList<TreeStream.Node>(ze.getChildren().size());
                for(ZipNode child:ze.getChildren()) l.add(new TreeStreamNode(zf,child));
                return l;
            }
            else return Collections.emptyList();
        }

        public InputStream getInputStream() {
            if(ze.hasEntry() && !ze.isDirectory()){
                try {
                    return zf.getInputStream(ze.getZipEntry());
                } catch (IOException e) {
                    log.log(Level.WARNING, "exception", e);
                    return null;
                }
            }
            return null;
        }

    }

    private void onDrag(DragGestureEvent dge, final DragSource ds) {
        Transferable t = getTransferable();
        if(t==null) return;
        log.finer("starting drag now");
        //NS 20051024: have noticed on Mac OS X 10.2.4/java1.5.0 sometimes drop doesn't succeed e.g. drop onto external window
        //and cursor changes back to normal. After this trags don't get initiated, e.g. call startDrag(..) and nothing happens
        //restarting app cures it..
        ds.startDrag (dge, DragSource.DefaultCopyDrop, t, new DragSourceListener() {
            public void dragEnter(DragSourceDragEvent dsde) {log.finest("dragenter");}
            public void dragOver(DragSourceDragEvent dsde) {log.finest("dragOver");}
            public void dropActionChanged(DragSourceDragEvent dsde) {log.finest("dropActionChanged"); }
            public void dragExit(DragSourceEvent dse) {log.finest("dragExit");}
            public void dragDropEnd(DragSourceDropEvent dsde) {log.finest("dragDropEnd");}
        });
    }

    private ArrayList<ZipEntry> getZipEntries(int[] sel, boolean includeDirectories, boolean recurse) {
        final ArrayList<ZipEntry> zes = new ArrayList<ZipEntry>(sel.length);
        for (int i = 0; i < sel.length; i++) {
            final ZipNode zipNode = (ZipNode) ct.getNode(sel[i]);
            addZipEntries(zipNode, includeDirectories, zes, recurse);
        }
        return zes;
    }

    private void addZipEntries(ZipNode zipNode, boolean includeDirectories, final ArrayList<ZipEntry> zes, boolean recurse) {
        if (zipNode.hasEntry()) {
            if(!includeDirectories && zipNode.isDirectory()) return;
            zes.add(zipNode.getZipEntry());
        }
        if(recurse){
            for(ZipNode child:zipNode.getChildren()){
                addZipEntries(child,includeDirectories,zes,recurse);
            }
        }
    }

    private void setupActions() {
        //copy the selected filenames to the clipboard
        final Action copyNamesAction = ActionHelper.setupAction(new AbstractAction("copynames") {
            public void actionPerformed(ActionEvent e) {
                List<ZipNode> l = getSelectedFiles(true, false, false);
                if (l.size()>0) {
                    StringBuffer sb = new StringBuffer();
                    for(ZipNode vf:l){
                        if(sb.length()>0) sb.append("\n");
                        sb.append(vf.getName());
                    }
                    try {
                        Clipboard cb = Toolkit.getDefaultToolkit().getSystemClipboard();
                        cb.setContents(
                                new StringSelection(sb.toString())
                                , new ClipboardOwner() { public void lostOwnership(Clipboard clipboard, Transferable contents) { }}
                        );
                    } catch (Exception e1) {
                        log.log(Level.WARNING, "couldn't copy "+sb.toString()+" to clipboard", e1);
                    }
                }

            }
        }, KeyStroke.getKeyStroke(KeyEvent.VK_C,Main.ctrl_mask),"copy");

        ct.registerAction("copynames",copyNamesAction);
        ct.registerInput(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT,KeyStroke.getKeyStroke(KeyEvent.VK_C,Main.ctrl_mask),"copynames");
        ct.registerInput(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT,KeyStroke.getKeyStroke(KeyEvent.VK_COPY,0),"copynames");


        new AutocompleteJumper(ct,new AutocompleteJumper.RowIdentifier() {
            public String getRowId(Object node) {
                return ((ZipNode)node).getName();
            }
        });
        //if we are viewing the root then tell hf to open url parent of file
        //todo: bug: this gets fired _after_ the currentnode has been changed,
        ct.addKeyListener(new KeyAdapter() {
            public void keyReleased(KeyEvent e) {
                if(e.getKeyCode()==KeyEvent.VK_LEFT){
                    if(file!=null){
                        ZipNode current = (ZipNode) ct.getNode();
                        if(null==model.getParent(current)){//we are at the root
                            File parent = file.getParentFile();
                            if(parent!=null){
                                try {
                                    hf.openUrl(Url.getUrl(parent));
                                } catch (CannotOpenUrlException e1) {
                                    log.log(Level.WARNING, "exception", e1);
                                }
                            }
                            else log.warning("couldn't open parent of file="+file+" because it was null");
                        }
                    }
                }
            }
        });
        ct.registerAction("mark",sml.getMarkAction());
        ct.registerInput(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT,KeyStroke.getKeyStroke(KeyEvent.VK_SPACE,0),"mark");
        ct.registerAction("clear",sml.getClearAction());
        ct.registerInput(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT,KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE,0),"clear");
        ct.registerAction("launch",EnterAction);
        ct.registerInput(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT,KeyStroke.getKeyStroke(KeyEvent.VK_ENTER,0),"launch");

        //right-click popup menu
        ct.addMouseListener(new MouseAdapter(){
            private void mouse(MouseEvent e) {
                if (e.getClickCount()==2){
                    EnterAction.actionPerformed(null);
                }
            }
            public void mousePressed(MouseEvent e) { mouse(e); }
            public void mouseReleased(MouseEvent e) { mouse(e); }
        });
    }
    private void writeZipFile() throws IOException {
        ZipOutputStream zos =new ZipOutputStream(new BufferedOutputStream(new FileOutputStream(file)));
        for(ZipEntry ze:root.getZipEntries()){
            zos.setMethod(ze.getMethod());
            zos.putNextEntry(ze);
        }
        zos.closeEntry();
        zos.close();
        final List<File> vfiles = Collections.emptyList();
        //notify listeners that the file has changed
        FileEventManager.getInstance().fireEvent(vfiles,vfiles,Collections.singletonList((file)));
    }
    private File writeTempFile(ZipEntry ze) throws IOException {
        return writeTempFiles(Collections.singletonList(ze)).get(0);
    }

    private List<File> getTempFiles(List<ZipEntry> zes) throws IOException {
        List<File> outs = new ArrayList<File>(zes.size());
        for (ZipEntry ze:zes) {
            File tempRoot= new File(System.getProperty("java.io.tmpdir"),"commanderbond_tempdata");
            File out = new File(tempRoot,ze.getName());
            outs.add(out);
        }
        return outs;
    }

    private List<File> writeTempFiles(List<ZipEntry> zes) throws IOException {
        //open the zipfile for each read - this may be inefficient, but saves issues arising from holding the file open while
        //we  do other things
        ZipFile zf = new ZipFile(file);
        List<File> outs = new ArrayList<File>(zes.size());
        for (ZipEntry ze:zes) {
            if(ze.isDirectory()) continue; //skip it
            InputStream is = zf.getInputStream(ze);
            File tempRoot= new File(System.getProperty("java.io.tmpdir"),"commanderbond_tempdata");
            File out = new File(tempRoot,ze.getName());
            File parent = out.getParentFile();
            if(parent!=null){
                boolean b=parent.mkdirs();
                if(!b) log.warning("couldn't create parent of out="+out+" parent="+parent);
            }
            else log.fine("parent of out="+out+" was null!");
            OutputStream os = new BufferedOutputStream(new FileOutputStream(out));
            byte[] buf = new byte[256];
            int i;
            while((i=is.read(buf))!=-1){
                os.write(buf,0,i);
            }
            is.close();
            os.close();
            outs.add(out);
        }
        zf.close();
        return outs;
    }

    private Action EnterAction = new AbstractAction() {
        private Url getArchiveUrl(String urlType, File node) throws MalformedURLException {
            Url url = Url.getUrl(urlType+":"+node.toURL().toExternalForm()+"!/");
            return url;
        }
        public void actionPerformed(ActionEvent e) {
            int[] sel = sml.getSelectedOrMarked(true, true,false);
            //weed out directories and nodes without zipentries
            final ArrayList<ZipEntry> l = getZipEntries(sel, false, false);

          //  List l = ct.getSelectedNodes(true,true);
            ZipEntry ze ;
            if(l.size()==1 && !(ze=l.get(0)).isDirectory()){  //ignore directories as enter causes them to open
                // if it's a jar/zip file then extract it to ${tmp} then browse it
                try {
                    if(ze.getName().toLowerCase().endsWith(".zip") ){
                        log.fine("got ourselves a nested zip");
                        File f = writeTempFile(ze);
                        Url url = getArchiveUrl("zip",f);
                        hf.openUrl(url);
                    }
                    else if(ze.getName().toLowerCase().endsWith(".jar")){
                        log.fine("got ourselves a nested jar");
                        File f = writeTempFile(ze);
                        log.fine("wrote tempfile="+f);
                        Url url = getArchiveUrl("jar",f);
                        log.fine("feeding url="+url+" to hf");
                        hf.openUrl(url);
                    }
                    else{
                        launchFiles(Collections.singletonList(ze));
                    }
                } catch (MalformedURLException e1) {
                    JOptionPane.showMessageDialog(ct,"Couldn't convert filename into URL\n" +
                            e1.getMessage(),"Open archive",JOptionPane.ERROR_MESSAGE);
                } catch (CannotOpenUrlException e1) {
                    log.log(Level.WARNING,"coudln't open archive",e1);
                    JOptionPane.showMessageDialog(ct,"Couldn't open archive\n" +
                            e1.getMessage(),"Open archive",JOptionPane.ERROR_MESSAGE);
                } catch (IOException e1) {
                    log.log(Level.WARNING,"coudln't write tempfile archive",e1);
                    JOptionPane.showMessageDialog(ct,"Couldn't extract file from archive\n" +
                            e1.getMessage(),"Open archive",JOptionPane.ERROR_MESSAGE);
                } catch(Exception e1){
                    log.log(Level.WARNING,"coudln't write tempfile archive",e1);
                    JOptionPane.showMessageDialog(ct,"Couldn't extract file from archive\n" +
                            e1.getMessage(),"Open archive",JOptionPane.ERROR_MESSAGE);
                }
                hf.getFocus();
            }
        }


    };


}
